
;*********************************************************************************
;* arith.s:
;*********************************************************************************

;********************************************************************************
;* This file is part of gibl:							*
;*	Gordons Interactive Basic Language                                      *
;********************************************************************************
;*    A Tiny Basic for the 6502 inspired by the NIBL Tiny Basic interpreter;	*
;*    "National Industrial Basic Language" originally for the INS8060 SC/MP	*
;*    system.									*
;*										*
;*    gibl is distributed under a "Source Available" license.			*
;*	It is NOT Open source and must not be treated as such.			*
;*										*
;*    See the file LICENSE for details.						*
;*										*
;*    gibl is Copyright (c) 2023 by Gordon Henderson				*
;********************************************************************************

	.include	"data.h"
	.include	"gibl.h"
	.include	"error.h"
	.include	"system.h"

	.include	"il.h"
	.include	"ilMacros.h"
	.include	"ilExec.h"
	.include	"ilUtils.h"

	.include	"arith.h"


;********************************************************************************
;* ADD:
;*	Take top 2 items off the stack, add them, put the result back. Simples
;*		s[1] := s[1] + s[0] ; sp--
;********************************************************************************

.proc	ADD
	ldx	arithPtr

	clc
	lda	arithStack-4,x
	adc	arithStack-2,x
	sta	arithStack-4,x

	lda	arithStack-3,x
	adc	arithStack-1,x
	sta	arithStack-3,x
	bvs	overflow

	dex
	dex
	stx	arithPtr

	rts
.endproc

.proc	overflow
	lda	#eVALU
	jmp	progErr
.endproc

;********************************************************************************
;* SUB:
;*	Take top 2 items off the stack, subtract them, put the result back.
;*		s[1] := s[1] - s[0] ; sp--
;********************************************************************************

.proc	SUB
	ldx	arithPtr

	sec
	lda	arithStack-4,x		; s[1] low
	sbc	arithStack-2,x		; s[0] low
	sta	arithStack-4,x

	lda	arithStack-3,x		; s[1] high
	sbc	arithStack-1,x		; s[0] high
	sta	arithStack-3,x
	bvs	overflow

	dex
	dex
	stx	arithPtr
	
	rts
.endproc

	
;********************************************************************************
;* mkPlus:
;*	Make the word at the current arithStack,x positive
;********************************************************************************

.proc	mkPlus
	sec
	lda	arithStack-2,x		; Low byte
	sbc	#1			; Subtract 1
	eor	#$FF			; ... and invert bits
	sta	arithStack-2,x

	lda	arithStack-1,x		; High byte
	sbc	#0
	eor	#$FF
	sta	arithStack-1,x

	iny
	rts
.endproc


;********************************************************************************
;* fixSigns:
;*	Record and fix (make positive) the signs of the top 2 numbers
;*	on the arith stack.
;********************************************************************************

.proc	fixSigns
	ldy	#0

	ldx	arithPtr
	lda	arithStack-1,x	; High
	bpl	:+		; ... is positive
	jsr	mkPlus

:	dex
	dex
	lda	arithStack-1,x	; High
	bpl	:+
	jsr	mkPlus

:	sty	num

	rts
.endproc


;********************************************************************************
;* MUL
;*	Take top 2 items off the stack, multiply them, put the result back.
;*		s[1] := s[1] * s[0] ; sp--
;*
;* Notes:
;*	Assumes regB and regC are consecutive in RAM. These are used to hold the
;*	32-bit result.
;*	Result is calculated directly on the stack.
;*	Algorithm is naive, but works well. Faster ones are out there...
;********************************************************************************

.proc	MUL

; Check signs and make positive if needed

	jsr	fixSigns

; Zero the 32-bit result in regB, regC:
;	Note: regB, regC must be consecutive in zero page.

	lda	#0		; Holds high byte of result
	sta	regB+0
	sta	regB+1
	sta	regB+2
	sta	regB+3

	ldx	arithPtr
	ldy	#15		; Loop counter

mulLoop:
	lsr	arithStack-1,x	; regB+1		; 7	Rotate right from the top byte down
	ror	arithStack-2,x	; regB+0		; 7	Bottom bit now in Carry
	bcc	mul2		; 2.5	No carry
				; 	... else add regA into the result
	pha			; 2	Temp save
	clc			; 2
	lda	arithStack-4,x	; regA+0		; 4
	adc	regB+2		; 4
	sta	regB+2		; 4

	pla			; 2	Restore top byte
	adc	arithStack-3,x	; regA+1		; 4

; Ripple A (top word) and regB/2,1,0 down to the right

mul2:	ror	a		; 2
	ror	regB+2		; 7
	ror	regB+1		; 7
	ror	regB+0		; 7

; Multply step...

	dey
	bne	mulLoop

; Last iteration

	ror	a		; 2
	ror	regB+2		; 7
	ror	regB+1		; 7
	ror	regB+0		; 7
	sta	regB+3		; Store top byte

; 32-bit result now in regB,regC
; .... copy bottom 16-bits to top of stack, discarding overflow
;	fixup the sign of the result and we're done.

	lda	regB+0		; Low byte
	sta	arithStack-4,x
	lda	regB+1		; High byte
	sta	arithStack-3,x

; Lets check for overflow ...

	lda	regB+2
	ora	regB+3
	beq	mulOK
	jmp	overflow

mulOK:
	dex
	dex
	stx	arithPtr

	lda	num
	and	#1
	bne	NEG
	rts
.endproc


;********************************************************************************
;* NEG:
;*	Negate the top item on the stack
;*		s[0] := - s[0]
;********************************************************************************

.proc	NEG
	ldx	arithPtr

	clc
	lda	arithStack-2,x	; s[0] low
	eor	#$FF		; ... invert
	adc	#1		; ... add 1
	sta	arithStack-2,x

	lda	arithStack-1,x	; s[0] high
	eor	#$FF		; ... Invert
	adc	#0		; ... add carry
	sta	arithStack-1,x

	rts
.endproc



;********************************************************************************
;* DIV
;*	Take top 2 items off the stack, divide them, put the result back.
;*		s[1] := s[1] / s[0] ; sp--
;*
;* Notes:
;*	Assumes regB and regC are consecutive in RAM. These are used to hold the
;*	32-bit result.
;*	Result is calculated directly on the stack.
;*	Algorithm is naive, but works well. Faster ones are out there...
;********************************************************************************


;********************************************************************************
; remDiv:
;	Common code for both the REM and DIV opcodes
;********************************************************************************

.macro	divStep
	.local	div2

	asl	arithStack-4,x	; regB+0		; Shift high bit of B into remainder
	rol	arithStack-3,x	; regB+1
	rol	regC+0
	rol	regC+1

	lda	regC+0
	sec
	sbc	arithStack-2,x	; regA+0
	sta	temp0
	lda	regC+1
	sbc	arithStack-1,x	; regA+1
	bcc	div2
	sta	regC+1
	lda	temp0
	sta	regC+0
	inc	arithStack-4,x	; regB+0
div2:
.endmacro

.proc	remDiv

; Check signs and make positive if needed

	jsr	fixSigns

; Check for div0

	ldx	arithPtr
	lda	arithStack-2,x
	ora	arithStack-1,x
	bne	divOK
	lda	#eDIV0
	jmp	progErr

divOK:

; regC is used to hold the remainder

	lda	#0
	sta	regC+0
	sta	regC+1
	ldx	arithPtr

.ifdef	UNROLL_DIV_LOOP
	.repeat	16
	  divStep
	.endrep
.else
	ldy	#16
:	  divStep
	dey
	bne	:-
.endif

; Result of division is in regB, the remainder in regC

	dex
	dex
	stx	arithPtr

	rts
.endproc


;********************************************************************************
; DIV
;	A := B / A
;********************************************************************************

.proc	DIV
	jsr	remDiv

; Check/Fix sign

	lda	num
	and	#1
	bne	NEG
	rts
.endproc


;********************************************************************************
; MOD:
;	A := B % A (remainder)
;********************************************************************************

.proc	MOD
	jsr	remDiv

; Remainder in regC, so

	ldx	arithPtr
	lda	regCL
	sta	arithStack-2,x
	lda	regCH
	sta	arithStack-1,x

; Check/Fix sign

	lda	num
	and	#1
	bne	NEG
	rts
.endproc


;********************************************************************************
;* getRnd
;*	Return the next pseudo random number
;********************************************************************************

.proc	getRnd
	clc
	lda	#17		; Any odd value
	adc	rndX
	sta	rndX
	adc	rndY
	sta	rndY
	asl	a
	adc	rndX
	tay
	eor	rndX

; Push into stack

	ldx	arithPtr
	sta	arithStack+0,x
	tya
	and	#$7F
	sta	arithStack+1,x
	inx
	inx
	stx	arithPtr
	rts
.endproc


;********************************************************************************
;* seedRnd
;*	Seed the PRNG..
;********************************************************************************

.proc	seedRnd
	lda	regAL
	sta	rndX
	lda	regAH
	sta	rndY
	rts
.endproc
